;;PS5MATCH.SCM

;MIT 6.001                                          Fall, 1997


;;                  PATTERN MATCHING: IMPLEMENTATION


(define (match pat dat)
  (cond ((or (is-atom? pat) (null? pat))
         (if (equal? pat dat)           ;must match exactly
             (list the-empty-dictionary)
             '()))                      ;no match: return empty list
         ((qmark-variable? pat)
          (match-qmark-variable pat dat))
         ((null? dat) (match-null pat))
         ((is-atom? dat) '())           ;FAIL: PAT is a pair
         (else (match-pairs pat dat)))) ;PAT & DAT are pairs


(define (match-qmark-variable qvar dat)
  (cond ((qmark? qvar) (list the-empty-dictionary))
        (((var-restriction qvar) dat)
         (list (make-singleton-dictionary qvar dat)))
            ;DAT does not satisfy var-restriction:
        (else '())))      ;no match


                 ;only a LIST OF TILDE-VARS or a QMARK-VAR matches null
(define (match-null pat)
  (cond ((and (list? pat)
              (for-all? pat tilde-variable?))
         (list
          (make-dictionary
           (map
            (lambda (tilde-var) (list tilde-var '()))
            (filter (lambda (var) (not (tilde? var))) pat)))))
        ((qmark-variable? pat) (match-qmark-variable pat '()))
        (else '())))                    ;FAIL


(define (match-pairs pat dat)           ;PAT & DAT are pairs
  (let ((pat0 (car pat)))
    (cond
     ((tilde? pat0)
      (match-initial-tilde (cdr pat) dat))
     ((tilde-variable? pat0)
      (match-initial-tildevar pat0 (cdr pat) dat))
     (else            ;PAT does not start with a tilde-variable
      (let ((dicts1 (match pat0 (car dat))))
        (if (null? dicts1)
            '()
            (all-pairs dicts1
                       (match (cdr pat) (cdr dat))
                       dict-merge)))))))

;  (all-pairs '(x1 ... xn) '(y1 ... ym) op) =
;       (list
;         (op x1 y1) (op x1 y2) ... (op x1 ym)   --|
;                             .                    |
;                             .                    |  ONE BIG LIST
;                             .                    |
;         (op xn y1) (op xn y2) ... (op xn ym))  --|


;(tensor '(x1 ... xn) '(y1 ... ym) op) =
;   (list
;     (list (op x1 y1) (op x1 y2) ... (op x1 ym))  --|
;                         .                          |
;                         .                          |List of Lists
;                         .                          |
;     (list (op xn y1) (op xn y2) ... (op xn ym))) --|

(define (tensor x-list y-list op)
    (map (lambda (x)
           (map (lambda (y) (op x y))
                y-list))
         x-list))


(define (all-pairs x-list y-list op)
  (apply append (tensor x-list y-list op)))


;; MATCH-INITIAL-TILDE
;To match (~ PAT1 PAT2 ..) against (DAT1 DAT2 ...)
;case 1: ~ matches nothing and
;          (PAT1 PAT2 ...) matches (DAT1 DAT2 ...)
;case 2: ~ matches against one or more of DAT1 DAT2 ....
;          This happens exactly when (~ PAT1 PAT2 ...)
;          matches (DAT2 ...).

(define (match-initial-tilde pat-after-tilde dat)
                        ;say PAT-AFTER-TILDE = (PAT1 PAT2 ...)
  (define (match-pat-aft~ dat) ;matches (~ PAT1 PAT2 ...) against DAT
    (append
     (match pat-after-tilde dat) ;~ not used to match anything
     (if (null? dat)
         '()                     ;already have dicts from null DAT
     ;check match when ~ is used to match a a sequence starting with DAT1:
         (match-pat-aft~ (cdr dat)))))
  (match-pat-aft~ dat))


;; MATCH-INITIAL-TILDEVAR
;To match (~VAR PAT1 PAT2 ..) against (DAT1 DAT2 ...)
;case 1: ~VAR matches nothing and (PAT1 PAT2 ...) matches (DAT1 DAT2 ...)
;case 2: ~VAR matches against one or more of DAT1 DAT2 ....
;        to see how,
;        (i)  match (~VAR PAT1 PAT2 ...) against (DAT2 ...) and then
;        (ii) put DAT1 at the start of the sequence that ~VAR matches
;             in each dictionary from (i)


(define (match-initial-tildevar tilde-var pat-after-tildevar dat)
  (let ((rproc (var-restriction tilde-var)))
         ;MATCH-PAT-AFT~VAR matches (TILDE-VAR PAT1 PAT2 ...) against DAT
    (define (match-pat-aft~var dat)
      (append
                                        ;dicts with TILDE-VAR matching empty
       (map (dictionary-reviser tilde-var '())
            (match pat-after-tildevar dat))
       (if (null? dat)
           '()
           (let ((dat1 (car dat)))
             (if
              (rproc dat1)   
;  Plan to match TILDE-VAR as (dat1 ...).  First match (TILDE-VAR pat)
;  against (CDR DAT), then update the value of TILDE-VAR to reflect that these
;  matches were done in context (DAT1 <>).  Namely, DAT1 must be CONS'd onto the
;  value of TILDE-VAR.
              (map                                
               (lambda (dict)
                 ((dictionary-reviser   
                   tilde-var
                   (cons dat1 (lookup tilde-var dict)))
                  dict))
               (match-pat-aft~var (cdr dat)))
              '())))))  ;fail if DAT1 does not satisfy TILDE-VAR's restriction
    (match-pat-aft~var dat)))

 

                             ;;DICTIONARIES

(define the-empty-dictionary '())

(define empty-dictionary? null?)

(define (make-dictionary bindings)
  bindings)

(define (lookup var dict)
  (if (empty-dictionary? dict)
      (error "attempt to LOOKUP empty dictionary" var)
      (let ((bind (assq var dict)))
        (if bind
            (cadr bind)
            (error "LOOKUP: undefined variable" var dict)))))


(define (make-singleton-dictionary var val)
  (list (list var val)))


(define (dictionary-reviser var val)
  (lambda (dict)
                ;revise/install value of VAR in a DICT to be VAL
        (cons (list var val)
              (filter
               (lambda (bind) (not (eq? var (car bind))))
               dict))))

(define dict-merge append)   ;ASSUMES NO REPEAT VARIABLES

;The effect of repeated variables can be achieved by restricting 
;rule applications to cases when the lookup values of the variables
;are EQUAL?.


                        ;;APPLYING RULES

;TEMPLATE = DATUM

;INSTANTIATE: TEMPLATE X DICTIONARY --> DATUM

(define (instantiate temp dict)
  (cond ((or (and (qmark-variable? temp)
                  (not (qmark? temp)))
             (and (tilde-variable? temp)
                  (not (tilde? temp))))
         (lookup temp dict))
        ((or (is-atom? temp) (null? temp)
                 ;when ? or ~ is in a template, treat it as an atom
             (qmark? temp) (tilde? temp))
         temp)
        (else
         ((if (tilde-variable? (car temp))
              append
              cons)
            (instantiate (car temp) dict)
            (instantiate (cdr temp) dict)))))


;                             RULES


(define (make-simple-rule pattern template)
  (list pattern template))

(define rule.pattern car)

(define rule.template cadr)

(define (simple? rule)          ;of the form (pattern template)
  (and (pair? rule)
       (pair? (cdr rule))
       (null? (cddr rule))      ;length is 2
       (not (procedure? (cadr rule)))))

(define (make-restricted-rule pattern template dict-pred)
  (list pattern template dict-pred))

(define rule.restriction caddr)

(define (restricted? rule)          ;the only rules of length > 2
  (pair? (cddr rule)))


(define (make-general-rule pattern result-procedure)
  (list pattern result-procedure))

(define (general? rule)          ;of the form (pattern restriction-proc)
  (and (pair? rule)
       (pair? (cdr rule))
       (procedure? (cadr rule))))

(define rule.result-procedure cadr)

;SOME-RULE-APPLICATION: RULE X DATUM --> (DATUM + FAIL)

         ;Do SOME APPLICATION of the rule to the ENTIRE DATUM
         ;(We choose the first dictionary of the match.)
(define (some-rule-application rule datum)
  (let ((dicts (match (rule.pattern rule) datum)))
    (if (null? dicts)
        (fail)
        (cond
         ((simple? rule)
	  (instantiate (rule.template rule) (car dicts)))

         ((restricted? rule)
          (let ((dict-ok? (rule.restriction rule)))
            (define (search maybe-ok-dicts)
              (if (null? maybe-ok-dicts)
                  (fail)
                  (let ((dict (car maybe-ok-dicts)))
                    (if (dict-ok? dict)
			(instantiate (rule.template rule) dict)
			(search (cdr maybe-ok-dicts))))))
            (search dicts)))

         ((general? rule)
          (let ((proc (rule.result-procedure rule)))
            (define (search maybe-ok-dicts)
              (if (null? maybe-ok-dicts)
                  (fail)
                  (let* ((dict (car maybe-ok-dicts))
                         (result (proc dict)))
                    (if (fail? result)
                        (search (cdr maybe-ok-dicts))                      
			result))))
            (search dicts)))

	 (else (error "unknown rule" rule))))))



;SOME-RULES-APPLICATION: LIST(RULE) X DATUM --> (DATUM + FAIL)

         ;Do some application of SOME RULE to the ENTIRE DATUM
         ;(We choose the first dictionary of the match of
         ; the first rule that has one.)
(define (some-rules-application rules datum)
  (if (null? rules)
      (fail)
      (let ((result (some-rule-application (car rules) datum)))
        (if (fail? result)
            (some-rules-application (cdr rules) datum)
            result))))


         ;Do some application of some rule to SOME SUB-DATUM
         ;(We choose the first dictionary of the match of
         ; the first rule that has one.)
(define (some-leftmost-outermost-application rules datum)
  (let ((result (some-rules-application rules datum)))
    (if (fail? result)
        (some-leftmost-subtrees-application rules datum)
        result)))
  

(define (some-leftmost-subtrees-application rules datum)
  (if (or (is-atom? datum) (null? datum))
      (fail)
      (let ((dat1     (car datum))
            (dat-rest (cdr datum)))
        (let ((dat1-result
               (some-leftmost-outermost-application
                rules
                dat1)))
          (if (fail? dat1-result)
              (let ((dat-rest-result
                     (some-leftmost-subtrees-application rules (cdr datum))))
                (if (fail? dat-rest-result)
                    (fail)
                    (cons dat1 dat-rest-result)))
              (cons dat1-result dat-rest))))))


             ;keep applying rules UNTIL NO RULE APPLIES ANYWHERE
(define (some-leftmost-outermost-final-form rules datum)
  (let ((result (some-leftmost-outermost-application rules datum)))
    (if (fail? result)
        datum
        (some-leftmost-outermost-final-form rules result))))


(define do-rules some-leftmost-outermost-final-form)



       ;;VARIABLES

(define (qmark? p) (eq? p '?))

(define (tilde? p) (eq? p '~))

(define (qmark-variable? p)
  (and (symbol? p)
       (char=? #\? (string-ref (symbol->string p) 0))))
         
(define (tilde-variable? p)
  (and (symbol? p)
       (char=? #\~ (string-ref (symbol->string p) 0))))


(define (var-restriction var)
  (let ((str (symbol->string var)))
    (if (and (> (string-length str) 1)
             (char=? #\% (string-ref str 1)))
        (let ((binding (assq var restriction-table)))
          (if binding
              (cadr binding)
              (error "VAR-RESTRICTION: no entry for" var)))
        constant-true-function)))


(define (constant-true-function pat) #t)


(define restriction-table
  (map list '(?%num ?%num1 ?%num2 ~%nums
              ?%not-num
              ?%not-num1
              ?%not-num2
              ?%var ?%var1 ?%var2
              ?%op
              ?%verb)
        (list number? number? number? number?
              (lambda (dat) (not (number? dat)))
              (lambda (dat) (not (number? dat)))
              (lambda (dat) (not (number? dat)))
              symbol? symbol? symbol?
              (lambda (dat) (assq dat ops-environment))
              (lambda (dat) (memq dat '(dance run sing cry laugh hate))))))

       ;;ATOMS

(define (is-atom? p)  ;must be matched exactly
  (or (number? p)
      (boolean? p)
      (and (symbol? p)
           (not (qmark-variable? p))
           (not (tilde-variable? p)))))

      ;;FAIL

(define fail-value (list  "FAIL"))

(define (fail) fail-value)

(define (fail? datum)
  (eq? datum fail-value))


            ;;STANDARD PROCEDURES

(define (filter p l)
  (cond ((null? l) '())
        ((p (car l))
         (cons (car l) (filter p (cdr l))))
        (else (filter p (cdr l)))))
      

(define (for-all? l p)
  (or (null? l)
      (and (p (car l)) (for-all? (cdr l) p))))



;TYPES

;MATCH: (PATTERN X DATUM) --> LIST(DICTIONARY)
;         returns dictionaries for all possible matches      

;DATUM = Tree(ATOM)
;ATOM = NUMBER + BOOLEAN + SYMBOL-WHICH-IS-NOT-A-VARIABLE

;PATTERN = Tree(ATOM + VARIABLE)  --- (i)  NO REPEATED VARIABLES
;                                     (ii) NOT SOLELY A TILDE-VARIABLE
;DICTIONARY = List(BINDING)
;BINDING = (QMARK-VARIABLE x DATUM) U
;          (TILDE-VARIABLE x List(DATUM))

;QMARK-VARIABLE   = Symbol starting with "?"
;TILDE-VARIABLE   = Symbol starting with "~"
