;;PS5RULES.SCM

; SYMBOLIC DERIVATIVES --- D(f) = df/dx

(define deriv-rules
                            ;D(F + G) = D(F) + D(G)
 (list
  (make-simple-rule
   '(deriv (?F + ?G) ?X)
   '((deriv ?F ?X) + (deriv ?G ?X)))
                            ;D(F * G) = (F * D(G)) + (G * D(F)) 
  (make-simple-rule
   '(deriv (?F * ?G) ?X)
   '((?F * (deriv ?G ?X))
     +
     (?G * (deriv ?F ?X))))
                            ;D(n) = 0
  (make-simple-rule
   '(deriv ?%num ?X)
   0)
                            ;D(y) = 1 when y = x
  (make-restricted-rule
   '(deriv ?%var ?X)
   1
   (lambda (dict)
     (eq? (lookup '?%var dict)             
          (lookup '?X dict))))
                            ;D(y) = 0 when y =\= x
  (make-restricted-rule
   '(deriv ?%var ?X)
   0
   (lambda (dict)
     (not (eq? (lookup '?%var dict)
               (lookup '?X dict)))))))

;(define test-deriv '(deriv ((a + b) * (2 + (a * a))) a))

;(do-rules deriv-rules test-deriv)
;Value: (((a + b) * (0 + ((a * 1) + (a * 1)))) + ((2 + (a * a)) * (1 + 0)))


;SIMPLIFYING INFIX BINARY ARITHMETIC

(define simplify-rules
 (list
  (make-simple-rule '(0 + ?F) '?F)

  (make-simple-rule '(0 * ?F) 0)

  (make-simple-rule '(1 * ?G) '?G)

  (make-general-rule   ;PRIMITIVE-OP
   '(?%num1 ?%op ?%num2)
   (lambda (dict)       ;RESULT PROCEDURE: DICT --> DATUM
     (let* ((opname (lookup '?%op dict))
            (opbinding (assq opname ops-environment)))
       (if opbinding
           ((cadr opbinding)
            (lookup '?%num1 dict)
            (lookup '?%num2 dict))
           (fail)))))    ;FAIL: UNKNOWN OPNAME

  (make-simple-rule  ;DISTRIBUTE-LEFT
   '(?%not-num * (?%not-num1 + ?%not-num2))
   '((?%not-num * ?%not-num1) + (?%not-num * ?%not-num2)))

  (make-simple-rule  ;DISTRIBUTE-RIGHT
   '((?%not-num1 + ?%not-num2) * ?%not-num)
   '((?%not-num1 * ?%not-num) + (?%not-num2 * ?%not-num)))

  (make-simple-rule    ;+NUMS-LEFT-RULE
   '(?%not-num + ?%num) '(?%num + ?%not-num))

  (make-simple-rule    ;*NUMS-LEFT-RULE
   '(?%not-num * ?%num) '(?%num * ?%not-num))

  (make-simple-rule    ;*ASSOC-LEFT
   '(?%not-num * (?G * ?H)) '((?%not-num * ?G) * ?H))

  (make-simple-rule    ;*num-ASSOC-LEFT
   '(?%num1 * (?%num2 * ?H)) '((?%num1 * ?%num2) * ?H))

  (make-simple-rule    ;+ASSOC-LEFT
   '(?%not-num + (?G + ?H)) '((?%not-num + ?G) + ?H))

  (make-simple-rule    ;+num-ASSOC-LEFT
   '(?%num1 + (?%num2 + ?H)) '((?%num1 + ?%num2) + ?H))

  (make-restricted-rule
   '(?%not-num1 + ?%not-num2) '(2 * ?%not-num1)
   (lambda (dict)
     (equal? (lookup '?%not-num1 dict)
             (lookup '?%not-num2 dict))))))


(define ops-environment
  (list
   (list '+ +)
   (list '- -)
   (list '* *)))

;(define test-exp '((1 * a) + (a + (b * 0))))

;(do-rules simplify-rules test-exp)
;Value: (2 * a)

;(do-rules (append deriv-rules simplify-rules) test-deriv)
;Value: ((2 + ((((2 * a) * a) + (b * a)) + (b * a))) + (a * a))


;;;DESUGARING QUOTES LISTS

;(do-rules quote-rules '(quote ((1 a) ((b) c))))


(define quote-rules
 (list
             ; '(E F_1 F_2 ...)  ==> (CONS 'E '(F_1 F_2 ...))
  (make-simple-rule
   '(quote (?E ~FF))
   '(cons (quote ?E) (quote (~FF))))

             ; '3  ==>  3       [3 is "SELF-EVALUATING"]
  (make-restricted-rule
   '(quote ?S) '?S
  (lambda (dict)
    (let ((S (lookup '?S dict)))
      (or (number? S)
          (boolean? S)))))

          ;;CONS Simplification:
             ; (CONS E '()) ==> (LIST E)
  (make-simple-rule
   '(cons ?E '()) '(list ?E))
  (make-simple-rule
   '(cons ?E (list ~FF)) '(list ?E ~FF))
  (make-simple-rule '(list) '())))

;(do-rules quote-rules '(quote ((1 a) ((b) c))))
;Value: (list (list 1 (quote a)) (list (list (quote b)) (quote c)))
